<?php

/**
 * Convert qt-php-like url into html-like url when urlrewrite is active<br>Works only on the path-part (and next parts) of the url<br>Ex: 'qtx_login.php' becomes 'login.html'
 * @param string $url
 * @param string $ext extension (with dot)
 * @param boolean|string $prefix file prefix (true for the default)
 * @return string the converted url, or the original url when urlrewrite is not active
 */
function Href(string $url, string $ext='.html', bool $prefix=true)
{
  if ( !QT_URLREWRITE ) return $url;
  if ( $prefix ) $prefix = APP.'_';
  if ( empty($url) ) die('Href: empty argument');
  // Ex: 'qtx_login.php?a=1' becomes 'login.html?a=1'
  $i = strlen($prefix);
  if ( $i && substr($url,0,$i)===$prefix ) $url = substr($url,$i); // $i>0
  if ( strpos($url,'.php') ) $url = str_replace('.php',$ext,$url);
  return $url;
}

/**
 * Returns the language path (with final /)
 * @param string $str
 * @return string
 */
function getLangDir(string $str='')
{
  if ( empty($str) ) {
    if ( defined('QT_LANG') ) return 'language/'.QT_LANG.'/'; //...
    global $oDB; $str = $oDB->getSetting('language','en');
  }
  return 'language/'.$str.'/';
}

function translate(string $file)
{
  if ( file_exists(getLangDir().$file) ) Return getLangDir().$file;
  return 'language/en/'.$file;
}

function useModule(string $name)
{
  return isset($_SESSION[QT]['module_'.$name]);
}

/**
 * Returns the translated word, or the lowercase or plural version, from the $L language dictionnary
 * @param string $k key to be searched in the dictonnary (dot in the key allows accessing sub-dictionnary)
 * @param bool|int $n number $n>1 indicates plural. False to skip the plural search and don't add number
 * @param bool $show_n adds the number before the word
 * @param string $sep_n character between the number and the word
 * @param array $A dictionnary (automatically assigned to full-dictionnary or sub-dictionnary)
 * @param string $pk parentkey (automatically assigned when accessing sub-dictionnary)
 * @return string (can be an array of all sub-items when 'key.*' is requested)
 */
function L(string $k='', $n=false, bool $show_n=true, string $sep_n=' ', array $A=[], string $pk='')
{
  if ( $n!==false && !is_int($n) ) throw new Exception('Invalid argument');
  // if ( empty(trim($k)) ) throw new Exception('Invalid argument');
  // initialise
  if ( empty($A) ) { global $L; $A=$L; } // sub-dictionnary (or full dictionnary $L)
  $str = $pk.$k; // default is the key (if the key cannot match a dictionnary entry)
  // check subarray
  if ( strpos($k,'.')>0 )
  {
    $part = explode('.', $k,2);
    if ( empty($A[$part[0]]) || !is_array($A[$part[0]]) ) return $str;
    if ( $part[1]==='*' ) return $A[$part[0]];
    return L($part[1], $n, $show_n, $sep_n, $A[$part[0]], $part[0].'.');
  }
  // check if plural can be used
  $s = $n==false || $n<2 ? '' : 's';
  // search word
  if ( !empty($A[$k.$s]) ) {
    $str = $A[$k.$s];
  } elseif ( !empty($A[ucfirst($k.$s)]) ) {
    $str = strtolower($A[ucfirst($k.$s)]);
  } elseif ( !empty($A[$k]) ) {
    $str = $A[$k];
  } elseif ( !empty($A[ucfirst($k)]) ) {
    $str = strtolower($A[ucfirst($k)]);
  } elseif ( trim($k)=='' ) {
    throw new Exception('Invalid argument');
  } else {
    if ( substr($str,0,2)==='E_' ) $str = 'error: '.substr($str,2); // key is an (un-translated) error code, returns the error code
    if ( strpos($str,'_')!==false ) $str = str_replace('_',' ',$str); // When word is missing, returns the key code without _ (with the parentkey if used)
    if ( isset($_SESSION['QTdebuglang']) && $_SESSION['QTdebuglang'] ) $str = '<span style="color:red">'.$str.'</span>';
  }
  // if show number
  return ($n!==false && $show_n ? $n.$sep_n : '').$str;
  // ABOUT KEYS:
  // A dot in the key indicates a sub-dictionnary entry (when words are stored in array of array)
  // Dictionnary have case-sensitive keys (and first character is uppercase) and can have language specific plural form (key+s)
  // Using lowercase-key will lowercase the translation
  // Example in french:
  //   L('Save') returns 'Sauver',
  //   L('save') returns 'sauver'
  // You can add a number $n as second argument to request the plural form of a word. When $n>1 the plural form is returned (if the plural form is not defined in the dictionnary, the singular translation is returned)
  // Example in french:
  //   L('Domain',0) returns '0 Domaine' - here no plural but value is inserted
  //   L('domain',1) returns '1 domaine' - same but lowercase version
  //   L('Domain',2) returns '2 Domaines' - returns the plural version (if plural version exists in the dictionnary)
  //   L('Domain',2,false) returns 'Domaines' - returns the plural version, while the number is not displayed
  // Sub-array level are accessible with the dot-key and can be returned as array with the '*':
  //   Example:
  //   L('DateMMM.1') returns 'January'
  //   L('DateMMM.*') returns ['January',...,'Decembre']
  // Fallback - If the requested key (or subkey) is not defined in the language file (or the sub-array is not defined) the function returns the key itself without '_'
  //   Example:
  //   L('Unknown_key') returns 'Unknown key'
  //   L('Error.404') returns 'Error.404'
  //   L('E_101') [key used as error code] returns 'error: 101' when not defined in the language file.
  // Debug - If you define a session variable 'QTdebuglang' set to '1', the function shows in red the key not defined in the language file. A session variable can be set with url 'index.php?debuglang=1'
}
function Lh(string $key='', $n=false, bool $show_n=true, string $sep_n=' ') { return QTattr(L($key, $n, $show_n, $sep_n)); }
function Lj(string $key='', $n=false, bool $show_n=true, string $sep_n=' ') { $words = L($key, $n, $show_n, $sep_n); return is_array($words) ? array_map( function($e){ return html_entity_decode($e, ENT_NOQUOTES); }, $words ) : html_entity_decode($words, ENT_NOQUOTES); }

/**
 * Same as ctype_digit (in case of ctype library is disabled on the server)
 * @param string $str
 * @return boolean
 */
function QTctype_digit(string $str)
{
  if ( function_exists('ctype_digit') ) return ctype_digit($str);
  if ( $str!=='' && preg_match('/^[0-9]+$/',$str) ) return true;
  return false;
}

/**
 * Assign GET/POST values into typed-variables listed in $vars<br>Prefix str: int: boo: or flo: defines the type (use str: if no prefix)<br>Suffix ! means a GET/POST is required
 * @param string $vars list of type:variable, space separated. Ex. "a! b int:c boo:d"
 * @param boolean $inGet read value from $_GET
 * @param boolean $inPost read value from $_POST
 * @param boolean $trim trim value
 * @param boolean $striptags remove tags
 * @return void global variables listed in $args are re-assigned if the value exists in GET/POST
 */
function QThttp(string $vars, bool $inGet=true, bool $inPost=true, bool $trim=true, bool $striptags=true)
{
  // NOTES:
  // When a user try to inject new variables, they are skipped (only variables in $vars are parsed from GET/POST).
  // It's recommended to initialise the variables before using this assigment function.
  // Supported types are only integer,float,boolean,string. In $vars they can be noted 'int:', 'flo:', 'boo:', 'str:' (missing type is casted as 'str:')
  // With type 'boo' (boolean), the variable is set to TRUE when GET/POST is '1' or 'true'. ALL OTHER values are evaluated as FALSE.
  // For required variable (suffix !), script stops if the variable is not in GET/POST (or if get/post as an empty string).
  // When values are not in the Http GET/POST (and not marked as required), the initial variable remains unchanged (can be a new variable with NULL value if the variable was not initialised).
  // When values are red from both GET and POST, the POST values are assigned after. POST-values overwrite GET-values, and POST can include other variables (in addition to those already assigned from GET)
  // For required variable, the script stops if the variable is declared in GET but as an empty string (even if the variable is also declared in POST)
  $vars=explode(' ',$vars);
  foreach($vars as $typedvar)
  {
    if ( empty($typedvar) ) continue;
    if ( strpos($typedvar,':')===false ) $typedvar = 'str:'.$typedvar; // default str: when no type
    $arr = explode(':',$typedvar); if ( count($arr)!==2 ) die('QThttp: invalid format');
    $type = substr(trim($arr[0]),0,3); // first 3-lettres defines the type ('boolean', 'bool', 'boo' are valid. 'bol' throws a data type error)
    $var = trim($arr[1]); // global variable name
    if ( substr($var,-1)=='!' ) { $var = substr($var,0,-1); $required=true; } else { $required=false; } // required becomes FALSE when a value exists in GET or POST
    global $$var;
    if ( $inGet && isset($_GET[$var]) ) {
      if ( $required && $_GET[$var]==='' ) die('QThttp: Required argument ['.$var.'] is without value'); // initially empty (before type check, trim or strip_tags)
      $required=false;
      if ( $trim ) $_GET[$var]=trim($_GET[$var]);
      if ( $striptags ) $_GET[$var]=strip_tags($_GET[$var]);
      if ( ($type==='int' || $type==='flo') && !is_numeric($_GET[$var]) ) die('QThttp: Invalid type for argument '.$var);
      switch($type) {
        case 'str': $$var = $_GET[$var]; break;
        case 'int': $$var = (int)$_GET[$var]; break;
        case 'boo': $$var = $_GET[$var]==='1' || strtolower($_GET[$var])==='true' ? true : false; break;
        case 'flo': $$var = (float)$_GET[$var]; break;
        default: die('QThttp: Invalid data type ['.$type.']');
      }
    }
    if ( $inPost && isset($_POST[$var]) ) {
      if ( $required && $_POST[$var]==='' ) die('QThttp: Required argument ['.$var.'] is without value'); // initially empty (before type check, trim or strip_tags)
      $required=false;
      if ( $trim ) $_POST[$var]=trim($_POST[$var]);
      if ( $striptags ) $_POST[$var]=strip_tags($_POST[$var]);
      if ( ($type==='int' || $type==='flo') && !is_numeric($_POST[$var]) ) die('QThttp: Invalid type for argument '.$var);
      switch($type) {
        case 'str': $$var = $_POST[$var]; break;
        case 'int': $$var = (int)$_POST[$var]; break;
        case 'boo': $$var = $_POST[$var]==='1' || strtolower($_POST[$var])==='true' ? true : false; break;
        case 'flo': $$var = (float)$_POST[$var]; break;
        default: die('QThttp: Invalid data type ['.$type.']');
      }
    }
    // Still required, if required but is not in GET nor in POST
    if ( $required ) die('QThttp: Required argument ['.$var.'] is missing');
  }
}

function QTasTag(array $arr, $current='', array $arrAttr=[], array $arrDisabled=[], string $tag='option', string $eol='', string $fx='', array $reject=[])
{
  if ( !empty($fx) && !function_exists($fx) ) die('QTasTag: requested function ['.$fx.'] is unknown');
  if ( is_int($current) ) $current = (string)$current;
  if ( !is_string($current) ) die('QTasTag: arg #2 must be int or string');
  // $arr indexes can be [int], they will be converted to [string]
  // When $arrDisabled is included, it must be an array of trimmed-strings
  // With $current, the operator == is required as $key will be [string] but $current can be [int]
  $str = '';
  foreach($arr as $key=>$value)
  {
    if ( in_array($key,$reject) ) continue;
    // format the value: $key and $value are converted to [string]
    $key = QTattr($key); // $key is alway used as [string] attribute hereafter
    if ( is_array($value) ) $value = reset($value);
    $value = isset($arrAttr['format']) ? sprintf($arrAttr['format'],$value) : (string)$value;
    if ( !empty($fx) ) $value = $fx($value);
    $strName = isset($arrAttr['name']) ? $arrAttr['name'] : '';
    $strClass = isset($arrAttr['class']) ? $arrAttr['class'] : '';
    if ( isset($arrAttr['classS']) && strlen($current)>0 && $current==$key ) $strClass = $arrAttr['classS'];
    if ( isset($arrAttr['current']) && isset($arrAttr['classC']) && $arrAttr['current']==$key ) $strClass = $arrAttr['classC'];
    $strStyle = isset($arrAttr['style']) ? $arrAttr['style'] : '';
    if ( isset($arrAttr['styleS']) && strlen($current)>0 && $current==$key ) $strStyle = $arrAttr['styleS'];
    if ( isset($arrAttr['current']) && isset($arrAttr['styleC']) && $arrAttr['current']==$key ) $strStyle = $arrAttr['styleC'];
    switch($tag)
    {
    // Attributes must use QTattr(). $key is already converted with QTattr()
    case 'option'  : $str .= '<option value="'.$key.'"'.(empty($strClass) ? '' : ' class="'.QTattr($strClass).'"').(empty($strStyle) ? '' : ' style="'.QTattr($strStyle).'"').($current==$key ? ' selected' : '').(in_array($key,$arrDisabled,true) ? ' disabled ': '').'>'.$value.'</option>'; break;
    case 'checkbox': $str .= '<input type="checkbox" value="'.$key.'"'.(empty($strClass) ? '' : ' class="'.QTattr($strClass).'"').(empty($strStyle) ? '' : ' style="'.QTattr($strStyle).'"').(empty($strName) ? '' : ' name="'.QTatttr($strName).'"').(in_array($key,$arrDisabled,true) ? ' disabled ': '').'/>'.$value; break;
    case 'hidden'  : $str .= '<input type="hidden" name="'.$key.'" value="'.QTattr($value).'"/>'; break;
    case 'span'    : $str .= '<span'.(empty($strClass) ? '' : ' class="'.QTattr($strClass).'"').'>'.$value.'</span>'.(isset($arrAttr['endline']) ? $arrAttr['endline'] : '' ); break;
    default: die('QTasTag: Invalid arg #4');
    }
    if ( !empty($eol) ) $str .= $eol;
  }
  return $str;
}

/**
 * Convert string to compliant Html-attribute/sql-value (trim and drop doublequote)
 * @param string $str
 * @param int $size maximum size (0 unchanged, can be negatif)
 * @param string $unquote doublequote replacement
 * @return string
 */
function QTattr(string $str, int $size=0, string $unquote='')
{
  $str = trim($str);
  if ( $str==='' ) return '';
  if ( $size && isset($str[$size]) ) $str = substr($str,0,$size); // negatif will drop final $size characters
  return strpos($str,'"')!==false ? trim(str_replace('"',$unquote,$str)) : $str;
}

/**
 * Add or change (or remove) a key-value in array. When value is NULL the key is removed
 * @param array $arr
 * @param integer|string $key
 * @param mixed $val
 * @return array (changed/added key-value is at the end)
 */
function QTarradd(array $arr, $key, $val)
{
  if ( !is_string($key) && !is_int($key) ) die('QTarradd: invalid argument(s)');
  if ( isset($arr[$key]) ) unset($arr[$key]); // remove the key
  if ( is_null($val) ) return $arr;
  $arr[$key] = $val; // add the key
  return $arr;
}
function QTarrget(array $arr,$subkey='title')
{
  // Converts an array of arrays into a simple array where the values are the [$subkey]element of each array (indexes are preserved).
  // When the [$subkey]element doesn't existing, the result includes NULL.
  // If one element of $arr is not an array, it REMAINS in the result. $subkey can be integer or string.
  foreach($arr as $k=>$a) {
    if ( !is_array($a) ) continue;
    $arr[$k] = isset($a[$subkey]) ? $a[$subkey] : null;
  }
  return $arr;
}

/**
 * Returns an array (key=>value) from a multfield string 'key1=value1;key2=value2'
 * @param string $str
 * @param string $sep separator
 * @param string $fx a function to apply to each value ex. 'trim', 'strtolower', 'urldecode'...
 * @return array (of string), keys are trimmed, unformated parts are skipped
 * <p>Returns an empty array when $str is empty (or no "=" or no keys)<br>
 * When duplicate keys exist, the last value overwrites previous values<br>
 * Values without keys are skipped, but a key without a value is valid i.e. 'key1=;key2=value2' returns ['key1'=>'','key2'=>'value2']</p>
 */
function QTexplode(string $str, string $sep=';', string $fx='')
{
  if ( empty($str) ) return array();
  if ( !empty($fx) && !function_exists($fx) ) die('QTexplode: '.$fx.' is unknown function');
  $arr = explode($sep,$str);
  $arrArgs = array();
  foreach($arr as $str)
  {
    if ( empty($str) || strpos($str,'=')===false ) continue; // skip parts without =
    $arrPart = explode('=',$str); $arrPart[0] = trim($arrPart[0]); // trim the keys
    if ( $arrPart[0]==='' ) continue; // skip when no-key
    $arrArgs[$arrPart[0]] = empty($fx) ? $arrPart[1] : $fx($arrPart[1]);
  }
  return $arrArgs;
}

/**
 * Explode the uri $str (or the current uri) to return an array of the query parts (not urldecoded by default)
 * @param string $str an uri like 'page.html?a=1&b=2#1' (if $str is empty, use the current URI)
 * @param string $fx ex: 'urldecode' each value (key remains unchanged)
 * @return array (empty array when the uri does not contains query parts)
 */
function QTexplodeUri(string $str='', string $fx='')
{
  if ( empty($str) ) $str = $_SERVER['REQUEST_URI'];
  $str = parse_url($str,PHP_URL_QUERY); // null if no url query part
  if ( empty($str) ) return array();
  if ( strpos($str,'&amp;')!==false ) $str = str_replace('&amp;','&',$str);
  return QTexplode($str,'&',$fx);
}

/**
 * Search a specific key-value in a multifield string 'a=1;b=2;c=3'
 * @param string $str
 * @param string $key
 * @param mixed $alt (value if key not found)
 * @param string $sep
 * @param string $fx
 * @return string can be also $alt [mixed]
 */
function QTexplodeGet(string $str, string $key, $alt='', string $sep=';', string $fx='')
{
  // QTexplodeGet('a=1;b=2', 'a') returns '1'
  if ( empty($str) ) die('QTexplodeGet Invalid multifield string');
  if ( empty($key) ) die('QTexplodeGet Invalid key');
  $arr = QTexplode($str,$sep,$fx); // can be [] when str is empty
  return isset($arr[$key]) ? $arr[$key] : $alt;
  // Note on alt:
  // for wrong format, $alt is returned. Ex: in 'a;b=2', a is not a declaration string, thus $alt is returned
  // a key without value is valid (set to ''). Ex: in 'a=;b=2', a is declared, thus '' is returned
}

/**
 * Implode an array of key-values into a ini-string "key1=value1&key2=value2"
 * @param array $arr
 * @param string $sep separator character
 * @param string $fx function to apply to each value (ex: urlencode or strtolower)
 * @param boolean $skipNull do not include key=value when value is null
 * @return string (or '' if $arr is empty)
 */
function QTimplode(array $arr, string $sep='&', string $fx='', bool $skipNull=true)
{
  if ( !empty($fx) && !function_exists($fx) ) die('QTimplode: requested function ['.$fx.'] is unknown');
  $str = '';
  foreach($arr as $key=>$value)
  {
    if ( $skipNull && is_null($value) ) continue;
    $str .= (isset($str[0]) ? $sep : '').$key.'='.(empty($fx) ? $value : $fx($value));
  }
  return $str;
}

/**
 * Switch between mail() and external-PHPMailer (unsing class.pop3.php and class.smtp.php)
 * @param string $strTo
 * @param string $strSubject
 * @param string $strMessage
 * @param string $strCharset
 * @param string $engine SMTP-engine (''=as configured, '1'=external-SMTP, '0'=php-mail)
 */
function QTmail(string $strTo, string $strSubject, string $strMessage, string $strCharset='utf-8', string $engine='')
{
  $strHeaders = 'Content-Type: text/plain; charset='.$strCharset;
  if ( $engine==='' ) $engine = $_SESSION[QT]['use_smtp'];
  switch($engine)
  {
  case '1':
    if ( substr($_SESSION[QT]['smtp_host'],0,4)=='pop3' )
    {
      require 'bin/class/class.pop3.php';
      $pop = new POP3();
      $pop->Authorise($_SESSION[QT]['smtp_host'], $_SESSION[QT]['smtp_port'], 30, $_SESSION[QT]['smtp_username'], $_SESSION[QT]['smtp_password'], 1);
    }
    $mail = new PHPMailer();
    $mail->IsSMTP(); // set mailer to use SMTP
    $mail->Host = $_SESSION[QT]['smtp_host'];  // specify main and backup server
    $mail->Port = $_SESSION[QT]['smtp_port'];  // set the SMTP port for the GMAIL server
    $mail->SMTPAuth = true;     // turn on SMTP authentication
    $mail->Username = $_SESSION[QT]['smtp_username']; // SMTP username
    $mail->Password = $_SESSION[QT]['smtp_password']; // SMTP password
    $mail->From = $_SESSION[QT]['admin_email'];
    $mail->FromName = '';
    $mail->AddAddress($strTo);
    $mail->AddReplyTo($_SESSION[QT]['admin_email']);
    $mail->IsHTML(false);  // set email format to HTML
    $mail->Subject = $strSubject;
    $mail->Body    = $strMessage;
    $mail->AltBody = $strMessage;
    if ( !$mail->Send() )
    {
      echo '<br>Message could not be sent.';
      echo '<br>Mailer Error: ' . $mail->ErrorInfo;
      echo '<br>Subject: '.$mail->Subject;
      echo '<br>Message: '.$mail->Body;
      exit;
    }
    //echo "Message has been sent";
    break;
  default:
    mail($strTo,$strSubject,$strMessage,'From:'.$_SESSION[QT]['admin_email']."\r\n".$strHeaders);
    break;
  }
}